
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <assert.h>

#include "utok.h"

static void *  pbuffer = NULL;

#define DEATH(...) { fprintf( stderr, __VA_ARGS__ ); exit( errno ); }
#define NODE_NAME		"/dev/usrdev"
#define	 BSIZE	4096

/*
 * Постранично нициализируем память размером в PAGE_COUNT
 * страниц символьными значениями начиная с символа 'A'.
 * Т.е. каждый байт первой страницы инициализирован
 * значением 'A', второй страницы 'B' и т.д.
 * */
static void init_buffer( char *  buffer )
{
	int i;
	unsigned long psize = (unsigned long) getpagesize();
	char	initch = 'A';
	for ( i = 0; i < PAGE_COUNT; ++i, ++initch ) {
		memset( (void*) buffer + i * psize, initch, psize );
	}
	return;
}

int main( int argc, char* argv[] ) {

	int fd = -1;
	ssize_t	count;
	char buffer [ BSIZE + 1 ];
	const unsigned long psize = (unsigned long) getpagesize();
	int retval = 0; // posix_memalign( (void**)&pbuffer, PAGE_COUNT * psize, psize );

	// Выделяем буфер размером PAGE_COUNT страниц, выравненный на размер страницы
	pbuffer = (void *) memalign( psize, PAGE_COUNT * psize );
	if ( pbuffer == NULL )
		DEATH( "Can't allocate memory with error: %d.", retval );

	assert( ( (unsigned long)pbuffer & (psize-1) ) == 0 );

	init_buffer( pbuffer );

	printf( "Start buffer %p.\n", pbuffer );

	if ( (fd = open( NODE_NAME, O_RDWR )) < 0 )
		DEATH( "Problem opening the node " NODE_NAME ": %s.\n", strerror( errno ) );

	// Передаём драйверу адрес инициализированого буфера
	ioctl( fd, USERDEV_IO_SET_ADDRESS, pbuffer );

	buffer[ BSIZE ] = '\0';

	// Выводим содержимое нашего же буфера, только уже прошедшего через
	// механизм direct I/O (т.е. так как наш буфер увидел драйвер)
	while ( ( count = read( fd, buffer, BSIZE ) ) > 0 ) {
		buffer[ count ] = '\0';
		printf( "%s\n", buffer );
	}

	ioctl( fd, USERDEV_IO_UNSET_ADDRESS );
	close( fd );

	free( (void*) pbuffer );
	return 0;
}
